import { Session } from "solid-start/session/sessions";
import {
  createClient,
  fetchExchange,
  errorExchange,
  makeOperation,
  Client,
  gql,
} from "urql";
import { authExchange } from "@urql/exchange-auth";
import { GraphQLError } from "graphql";
import z, { Infer } from "myzod";

const getHost = () => z.string().parse(process.env.DIRECTUS_HOST);
const getPort = () =>
  z
    .bigint()
    .parse(
      process.env.DIRECTUS_PORT && Number.parseInt(process.env.DIRECTUS_PORT)
    );
const getGraphqlUrl = () => `http://${getHost()}:${getPort()}/graphql`;

export const createDirectusSystemClient = () => {
  const client = createClient({
    url: getGraphqlUrl() + "/system",
  });

  return client;
};

export const createDirectusClient = (
  session: Session,
  onAuthError?: Function
) => {
  const client = createClient({
    url: getGraphqlUrl(),
    exchanges: [
      authExchange({
        getAuth: async ({ authState }) => {
          if (!authState) {
            const token = session.get("accessToken");
            const refreshToken = session.get("refreshToken");
            if (token && refreshToken) {
              return { token, refreshToken };
            }
            return null;
          }

          return null;
        },
        addAuthToOperation: ({ authState, operation }: any) => {
          if (!authState || !authState.token) {
            return operation;
          }

          const fetchOptions =
            typeof operation.context.fetchOptions === "function"
              ? operation.context.fetchOptions()
              : operation.context.fetchOptions || {};

          return makeOperation(operation.kind, operation, {
            ...operation.context,
            fetchOptions: {
              ...fetchOptions,
              headers: {
                ...fetchOptions.headers,
                Authorization: `Bearer ${authState.token}`,
              },
            },
          });
        },
      }),
      errorExchange({
        onError: async (error, o) => {
          console.log("code", error.graphQLErrors[0].extensions.code);
          console.log(error.graphQLErrors[0].extensions.graphqlErrors);
          const isAuthError = error.graphQLErrors.some(
            (e: GraphQLError) =>
              ["TOKEN_EXPIRED", "FORBIDDEN"].indexOf(
                e.extensions?.code as string
              ) >= 0
          );
          console.log("isAuthError", isAuthError);
          if (onAuthError && isAuthError) {
            onAuthError();
          }
        },
      }),
      fetchExchange,
    ],
  });

  return client;
};

export const authSchema = z.object(
  {
    access_token: z.string(),
    refresh_token: z.string(),
    expires: z.number(),
  },
  { allowUnknown: true }
);

export type AuthSchema = Infer<typeof authSchema>;

export const login = async (options: {
  username: string;
  password: string;
  client?: Client;
}) => {
  const client = options.client || createDirectusSystemClient();

  const result = await client
    .mutation(
      gql`
    mutation {
      auth_login(email: "${options.username}", password: "${options.password}") {
        access_token
        refresh_token
        expires
      }
    }`
    )
    .toPromise();

  if (result.error) {
    throw result.error;
  }

  return authSchema.parse(result.data.auth_login);
};

export const refresh = async (options: {
  refreshToken: string;
  client?: Client;
}) => {
  const client = options.client || createDirectusSystemClient();

  const result = await client
    .mutation(
      gql`
    mutation {
      auth_refresh(refresh_token: "${options.refreshToken}") {
        access_token
        refresh_token
        expires
      }
    }`
    )
    .toPromise();

  if (result.error) {
    throw new Error(result.error.toString());
  }

  /*
  console.log("code", result.error.graphQLErrors[0].extensions.code);
  console.log(result.error.graphQLErrors[0].extensions.graphqlErrors);
  console.dir(result.error);
  throw new Error(result.error.response);
  */

  return authSchema.parse(result.data.auth_refresh);
};

export const logout = async (options: {
  refreshToken: string;
  client?: Client;
}) => {
  const client = options.client || createDirectusSystemClient();

  const result = await client
    .mutation(
      gql`
    mutation {
      auth_logout(refresh_token: "${options.refreshToken}")
    }`
    )
    .toPromise();

  if (result.error) {
    throw result.error;
  }

  return z.boolean().parse(result.data.auth_logout);
};
