import { Result } from "postcss";
import { useNavigate } from "solid-app-router";
import {
  createEffect,
  resetErrorBoundaries,
  ResourceReturn,
  sharedConfig,
  startTransition,
} from "solid-js";
import { createStore, produce } from "solid-js/store";
import { isRedirectResponse, LocationHeader } from "solid-start/server";

// Should handle redirects, but it broke "quickly"
export const useServer = function (
  serverFunction: Function,
  callback?: Function
) {
  const navigate = useNavigate();

  return async () => {
    try {
      const res = await serverFunction();
      return callback ? callback(res) : res;
    } catch (err) {
      if (!(err instanceof Response)) {
        throw err;
      }

      if (!isRedirectResponse(err)) {
        return callback ? callback(err) : err;
      }

      const newLocation = err.headers.get(LocationHeader);
      if (!newLocation) {
        throw new Error("Got RedirectResponse without location");
      }

      startTransition(function () {
        navigate(newLocation);
        resetErrorBoundaries();
      });

      return null;
    }
  };
};

export const awaitJson = <T extends (...args: any[]) => any>(fn: T) => {
  return async (...args: Parameters<T>) => {
    const result = await fn(...args);

    if (!result) {
      return result;
    }

    if (!(result instanceof Response)) {
      return result;
    }

    const contentType = result.headers.get("content-type");
    if (!contentType) {
      return result;
    }

    if (contentType.indexOf("application/json") === -1) {
      return result;
    }

    const jsonContent = await result.json();

    return jsonContent;
  };
};

export const composeRouteData =
  (functionMap: Record<any, (state: any) => ResourceReturn<any>>) => () => {
    const [state, setState] = createStore<Record<string, any>>({});

    const repos = {} as Record<string, ResourceReturn<any>>;

    for (const dataKey of Object.keys(functionMap)) {
      repos[dataKey] = functionMap[dataKey](state);

      setState(dataKey, repos[dataKey][0]());

      createEffect(function () {
        const newValue = repos[dataKey][0]();
        setState(
          produce(function (s) {
            s[dataKey] = newValue;
          })
        );
      });
    }

    const refetch = function () {
      for (const dataKey of Object.keys(functionMap)) {
        repos[dataKey][1].refetch();
      }
    };

    return [state, refetch, repos];
  };

export type RC = {
  request: Request;
  context: Record<any, any>;
  responseHeaders: Headers;
  url: URL;
};
export const createRC = function (
  request?: Request,
  responseHeaders?: Headers
) {
  const context = (sharedConfig?.context as any)?.requestContext;

  if (!request && context) {
    request = context.request;
  }

  if (!responseHeaders && context) {
    responseHeaders = context.responseHeaders;
  }

  if (!request || !responseHeaders) {
    throw new Error("Invalid request context");
  }

  const rc: RC = {
    request,
    responseHeaders,
    url: new URL(request.url),
    context: {},
  };

  return rc;
};
export const withRC = function <
  R extends any[],
  T extends (rc: RC, ...args: R) => any
>(fn: T) {
  return async function (
    this: any,
    ...args: R
  ): Promise<ReturnType<Awaited<T>>> {
    try {
      return await fn(createRC(this.request, this.responseHeaders), ...args);
    } catch (err) {
      if (!(err instanceof Response)) {
        console.log(err);
      }

      throw err;
    }
  };
};
